{ Knihovna funkci pro manipulaci se stringy

  Autor:  Gord
  Verze 1.5.1
    - (2.10.2013)zmena typu parametru funkce Base64ToBitmap (kvuli prechodu na novy GG)
    - (15.12.2013) oprava DecToHex pro nektera zaporna ciska vypisuje 0000000
    - (15.12.2013) nova funkce FetchFight()
    - (11.8.2021)  nova funkce HtmlEntityDecDecode()

  ----------------
  Funkce verze:  1.0.0
    StringCompress(str:string):string;
      - vyhaze mezery, prevede na mala pismena, prvni znak slova vzdy velky
      - ekvivalent FilenameFromText z knihovny RelToAbsPath

    RemoveInvalidPathChars(str:string):string;
      - odstrani znaky, ktere mohou delat potize ve jmenech souboru

    IntToStrLength(iNum,len:integer):string;
      - vypis Int na zadany pocet znaku s levostrannymi nulami

    Base64ToBitmap(str:string):TBitMap;
      - vytvoreni bitove mapy z Base64 stringu (uvolnit bitmapu musi volajici)

  Funkce verze 1.1.0
    StrToStrings(str,separator:string):TStrings;
      - rozdeli string podle oddelovace 'separator' a prevede do TStrings

    StringsToStr(a:sTStrings; separator:string):string;
      - prevede TStrings na string oddeleny 'separatorem'

    GetCurrentPluginVersion:string;
      - vraci cislo verze pluginu nastavene v *.meta.ini souboru proave spusteneho
        pluginu

  Funkce verze 1.2.0
    function HexCharToInt(HexChar: String): Integer;
    function DecToHex(iNum,len:integer):string;
    function HexToDec(sHex: String): Integer;

  Funkce verze 1.3.0
    function AutoWrapText(text:string; len:integer):string;

  Funkce verze 1.4.0 - prace s CSV souborem
    procdure LibCsvInit();
      - inicializace bufferu pro cteni radku a polozek z CSV souboru
    procedure LibCsvFree();
      - uvolni pamet, ...
    function LibCsvReadLine(fStream:TFileStream):string;
      - nacte ze streamu radek
    function LibCsvParseRow(row:string);
      - naplni LibCsvItems polozkami z radku CSV souboru

  Funkce verze 1.5.0
    fuction HtmlEntityDecDecode(text:string);
      - prevede entity s dekadickym cislem (napr. &#225 je dlouhe a), ktere
        funkce HtmlEntityDecode() nedekoduje
}

//odstraneni znaku, ktere mohou delat potize ve jmenech souboru
function RemoveInvalidPathChars(str:string):string;
begin
  Result:=RegexReplace('[~^/|;,*!?@&\"''<>=+{}]',str,'',false);
  //dvojtecka v ceste, pokud neni za pismenem disku, je chybna a mela by jit pryc
end;

//vyhaze mezery, prevede na mala pismena, prvni znak slova vzdy velky
function StringCompress(str:string):string;
var VelkePismenoZa:string;
    s:string;
    i,v:integer;
begin
  VelkePismenoZa:=' `~!?@%^&*+=:;,"<>''|/\';
  Result:='';
  v:=1;
  str:=LowerCase(str);
  for i:=1 to length(str) do
  begin
    s:=copy(str,i,1);
    if(pos(s,VelkePismenoZa)>0) then v:=1
    else begin
      if(v=1) then Result:=Result+UpperCase(s)
      else Result:=Result+s;
      v:=0;
    end;
  end;
  Result:=Result;
end;

//vypis Int na zadany pocet znaku s levostrannymi nulami
function IntToStrLength(iNum,len:integer):string;
begin
  Result:=IntToStr(iNum);
  if(length(Result)<len) then begin
    Result:=copy('0000000000000000000000000000000',1,len-length(Result))+Result;
  end;
end;

//vytvoreni bitove mapy z Base64 stringu (uvolnit bitmapu musi volajici)
function Base64ToBitmap(str:ansistring):TBitMap;
var s:ansistring;
    stream:TMemoryStream;
begin
  stream:=TMemoryStream.Create;
  try
    s:=DecodeBase64(str);
    stream.Write(s,Length(s));
    stream.Position:=0;
    Result:=TBitMap.Create();
    Result.LoadFromStream(stream);
  finally
    stream.free;
  end;
end;

function StrToStrings(str,separator:string):TStrings;
begin
  Result:=TStringList.Create();
  while (str<>'') do begin
    Result.Add(Fetch(str,separator));
  end;
end;

function StringsToStr(a:TStrings;separator:string):string;
var i:integer;
begin
  Result:='';
  for i:=0 to a.Count-1 do begin
    if(i>0) then Result:=Result+separator;
    Result:=Result+a[i];
  end;
end;

function GetCurrentPluginVersion:string;
begin
  Result:=GeoGetPackageVersion(ReplaceString(GEOGET_SCRIPTFULLNAME,'.ggp.pas','.meta.ini'));
end;

function HexCharToInt(HexChar: String): Integer;  {prevod jednoho Hex znaku na Int}
begin
  Result:=0;
  HexChar:=Uppercase(HexChar);
  if((HexChar>='0') and (HexChar<='9')) then Result:=Ord(HexChar[1])-48
  else if((HexChar>='A') and (HexChar<='Z')) then Result:=Ord(HexChar[1])-65+10;
end;

function HexToDec(sHex: String): Integer;   {prevod Hex na Dec}
var i:integer;
begin
  Result:=0;
  for i:=1 to Length(sHex) do begin
    Result:=Result+(HexCharToInt(copy(sHex,i,1))*Trunc(Power(16,Length(sHex)-i)));
  end;
end;

function DecToHex(iNum,len:integer):string;
var iByte:byte;
    s:string;
begin
  s:='0123456789ABCDEF';
  Result:='';
  while(iNum<>0) do begin
    iByte:=iNum and $0f;
    iNum:=iNum shr 4;
    Result:=s[iByte+1]+Result;
  end;
  if(length(Result)<len) then Result:=copy('00000000',1,len-length(Result))+Result
  else if(length(Result)>len) then Result:=copy(Result,1,len);
end;

function AutoWrapText(text:string;len:integer):string;
var reg:string;
begin
  //zalomi text na bilem znaku nejblize k zadane delce
  reg:='(.{0,'+IntToStr(len)+'}[^\s]*)\s?';
  Result:=RegexReplace(reg,trim(text),'$1'+CRLF,true);
  //Result:=RegexReplace(CRLF+'$',Result,'',true);
end;

var rLibCsv:record
    iBufferLength:integer;
    fStream:TFileStream;
    cDelimiter:char;   //oddelovac polozek, implicitne carka
    sRest:ansistring;   //nacteny zbytek - zacatek dalsiho radku
    //sBuffer2:ansistring; //
    aItems:TStringList; //parsovane polozky z radku
    end;
procedure LibCsvInit(fStream:TFileStream); //inicializace bufferu pro cteni radku a polozek z CSV souboru
begin
  rLibCsv.iBufferLength:=10000;
  rLibCsv.fStream:=fStream;
  rLibCsv.sRest:='';
  rLibCsv.aItems:=TStringList.Create();
  //(script enine to nepodporuje)
  //rLibCsv.aItems.DelimiterText:=',';
  //rLibCsv.aItems.QuoteChar:='"';
end;

procedure LibCsvFree; //uvolni pamet, ...
begin
  rLibCsv.aItems.Free();
  rLibCsv.aItems:=nil;
  rLibCsv.sRest:='';
  rLibCsv.fStream:=nil;
end;

function LibCsvParseRow(row:string):TStringList;//naplni LibCsv.aItems polozkami z radku CSV souboru a vrati je
begin
  rLibCsv.aItems.CommaText:=row;
  //nebo lepe (ale script enine to nepodporuje)
  //rLibCsv.aItems.DelimitedText:=row;
  Result:=rLibCsv.aItems;
end;

function LibCsvIsRowSeparator(sStr:ansistring; var bLFInItem:boolean):integer;
var iLen,nUvozovky:integer;
begin
  {
  Result:=Pos(#10,sStr);
  }
  //### tady by bylo potreba osetrit konce radku uprostred polozky, tedy
  //### mezi uvozovkami, ale tohle neni odladene a cteni gcm katalogu CWG
  //### v CwgList2 se zacykli (nebo je to tak pomale proti Pos() funkci,
  //### ze jsem nevydrzel cekat?
  iLen:=Length(sStr);
  nUvozovky:=0;
  bLfInItem:=false;
  for Result:=1 to iLen do begin
    if(sStr[Result]='"') then Inc(nUvozovky);
    if(sStr[Result]=#10) then begin
      //pokud je pred LF lichy pocet uvozovek, jsme uprostred polozky a neni to konec radku
      if((nUvozovky mod 2)=0) then exit;
      bLfInItem:=true;
    end;
  end;
  Result:=-1;
end;

function LibCsvReadLine:string; //nacte ze streamu radek
var sAnsi:ansistring;
    bEnd,bLfInclude:boolean;
    i:integer;
begin
  Result:='';
  bEnd:=rLibCsv.fStream.Position>=rLibCsv.fStream.Size;
//{$ifdef DEBUG_HELPER} LDH('LibCsvReadLine(): start, bEnd='+IntToStr(bEnd as integer)); {$endif}
  while(true) do begin
    //otestujeme, zda uz je nacten konec radku
    i:=LibCsvIsRowSeparator(rLibCsv.sRest,bLfInclude);
    if(i>0) then begin //pokud je nacten konec radku, vratime radek
      Result:=Trim(Copy(rLibCsv.sRest,1,i));
//{$ifdef DEBUG_HELPER} LDH('LibCsvReadLine(): Obsahuje nejaka polozka radku odradkovani? '+IntToStr(bLfInclude as integer)); {$endif}
//{$ifdef DEBUG_HELPER} LDH('LibCsvReadLine(): Vraceny radek: '+Result); {$endif}
      rLibCsv.sRest:=Copy(rLibCsv.sRest,i+1,Length(rLibCsv.sRest));
      SetLength(sAnsi,0);
      exit;
    end;
    if(bEnd) then begin
{$ifdef DEBUG_HELPER} LDH('LibCsvReadLine(): docten konec souboru, i='+IntToStr(i)
  +', sRest='+rLibCsv.sRest); {$endif}
      break;
    end;
    SetLength(sAnsi,rLibCsv.iBufferLength);
    i:=rLibCsv.fStream.Read(sAnsi,rLibCsv.iBufferLength);  //pokud jsme na konci, rovnou navrat
    if(i<>rLibCsv.iBufferLength) then begin
      SetLength(sAnsi,i); //zkratime na skutecne nacteny pocet znaku
      bEnd:=true;
    end;
    rLibCsv.sRest:=rLibCsv.sRest+sAnsi;
  end;
  //sem se dostaneme jen kdyz neni nacten konec radku na konci souboru
  Result:=rLibCsv.sRest;
  rLibCsv.sRest:='';
  SetLength(sAnsi,0);
end;

function BomUtf8:string;
begin
  Result:=chr(239)+chr(187)+chr(191);
end;

function HtmlEntityDecDecode(str:string):string;
  //prevede entity s dekadickym cislem (napr. &#225 je dlouhe a), ktere
  //funkce HtmlEntityDecode() nedekoduje
  //seznam: treba https://www.obkb.com/dcljr/charstxt.html
var aEntity: array[0..255] of string;
    aChar: array[0..255] of string;
    i:integer;
    bUtf:boolean;
begin
  aEntity[168]:='&#168;'; aChar[168]:=''; //umlaut, dvojtecka nad pismenem
  aEntity[169]:='&#169;'; aChar[169]:='';
  //aEntity[170]:='&#170;'; aChar[170]:='';
  //aEntity[171]:='&#171;'; aChar[171]:='';
  //aEntity[172]:='&#172;'; aChar[172]:='';
  //aEntity[173]:='&#173;'; aChar[173]:='';
  aEntity[174]:='&#174;'; aChar[174]:='';
  //aEntity[175]:='&#175;'; aChar[175]:='';
  aEntity[176]:='&#176;'; aChar[176]:=''; //stupen
  aEntity[177]:='&#177;'; aChar[177]:='';
  //aEntity[178]:='&#178;'; aChar[178]:='';
  //aEntity[179]:='&#179;'; aChar[179]:='';
  aEntity[180]:='&#180;'; aChar[180]:=''; //carka nad pismenem
  aEntity[181]:='&#181;'; aChar[181]:=''; //micro
  //aEntity[182]:='&#182;'; aChar[182]:='';
  aEntity[183]:='&#183;'; aChar[183]:=''; //tecka uprostred radku
  aEntity[184]:='&#184;'; aChar[184]:=''; //cedilla, ocasek pod pismenem
  //aEntity[185]:='&#185;'; aChar[185]:='';
  //aEntity[186]:='&#186;'; aChar[186]:='';
  //aEntity[187]:='&#187;'; aChar[187]:='';
  aEntity[188]:='&#188;'; aChar[188]:='1'; //ctvrtina
  aEntity[189]:='&#189;'; aChar[189]:='1'; //polovina
  aEntity[190]:='&#190;'; aChar[190]:='3'; //tri ctvrtiny
  aEntity[191]:='&#191;'; aChar[191]:='?'; //obraceny otaznik
  aEntity[192]:='&#192;'; aChar[192]:='A';
  aEntity[193]:='&#193;'; aChar[193]:='';
  aEntity[194]:='&#194;'; aChar[194]:='';
  aEntity[195]:='&#195;'; aChar[195]:='A';
  aEntity[196]:='&#196;'; aChar[196]:='';
  aEntity[197]:='&#197;'; aChar[197]:='A';
  aEntity[198]:='&#198;'; aChar[198]:='A';
  aEntity[199]:='&#199;'; aChar[199]:='';
  aEntity[200]:='&#200;'; aChar[200]:='E';
  aEntity[201]:='&#201;'; aChar[201]:='';
  aEntity[202]:='&#202;'; aChar[202]:='E';
  aEntity[203]:='&#203;'; aChar[203]:='';
  aEntity[204]:='&#204;'; aChar[204]:='I';
  aEntity[205]:='&#205;'; aChar[205]:='';
  aEntity[206]:='&#206;'; aChar[206]:='';
  aEntity[207]:='&#207;'; aChar[207]:='I';
  aEntity[208]:='&#208;'; aChar[208]:='?';
  aEntity[209]:='&#209;'; aChar[209]:='N';
  aEntity[210]:='&#210;'; aChar[210]:='O';
  aEntity[211]:='&#211;'; aChar[211]:='';
  aEntity[212]:='&#212;'; aChar[212]:='';
  aEntity[213]:='&#213;'; aChar[213]:='O';
  aEntity[214]:='&#214;'; aChar[214]:='';
  aEntity[215]:='&#215;'; aChar[215]:=''; //krat
  aEntity[216]:='&#216;'; aChar[216]:='O'; //preskrtnute O
  aEntity[217]:='&#217;'; aChar[217]:='U';
  aEntity[218]:='&#218;'; aChar[218]:='';
  aEntity[219]:='&#219;'; aChar[219]:='U';
  aEntity[220]:='&#220;'; aChar[220]:='';
  aEntity[221]:='&#221;'; aChar[221]:='';
  //aEntity[222]:='&#222;'; aChar[222]:='';
  //aEntity[223]:='&#223;'; aChar[223]:='';
  aEntity[224]:='&#224;'; aChar[224]:='a';
  aEntity[225]:='&#225;'; aChar[225]:='';
  aEntity[226]:='&#226;'; aChar[226]:='';
  aEntity[227]:='&#227;'; aChar[227]:='a';
  aEntity[228]:='&#228;'; aChar[228]:='';
  aEntity[229]:='&#229;'; aChar[229]:='a';
  aEntity[230]:='&#230;'; aChar[230]:='a';
  aEntity[231]:='&#231;'; aChar[231]:='';
  aEntity[232]:='&#232;'; aChar[232]:='e';
  aEntity[233]:='&#233;'; aChar[233]:='';
  aEntity[234]:='&#234;'; aChar[234]:='e';
  aEntity[235]:='&#235;'; aChar[245]:='';
  aEntity[236]:='&#236;'; aChar[236]:='i';
  aEntity[237]:='&#237;'; aChar[237]:='';
  aEntity[238]:='&#238;'; aChar[238]:='';
  aEntity[239]:='&#239;'; aChar[239]:='i';
  //aEntity[240]:='&#240;'; aChar[240]:='';
  aEntity[241]:='&#241;'; aChar[241]:='n';
  aEntity[242]:='&#242;'; aChar[242]:='o';
  aEntity[243]:='&#243;'; aChar[243]:='';
  aEntity[244]:='&#244;'; aChar[244]:='';
  aEntity[245]:='&#245;'; aChar[245]:='o';
  aEntity[246]:='&#246;'; aChar[246]:='';
  aEntity[247]:='&#247;'; aChar[247]:='';
  aEntity[248]:='&#248;'; aChar[248]:='o';
  aEntity[249]:='&#249;'; aChar[249]:='u';
  aEntity[250]:='&#250;'; aChar[250]:='';
  aEntity[251]:='&#241;'; aChar[251]:='u';
  aEntity[252]:='&#252;'; aChar[252]:='';
  aEntity[253]:='&#253;'; aChar[253]:='';
  //aEntity[254]:='&#254;'; aChar[170]:='';
  aEntity[255]:='&#255;'; aChar[255]:='y';
  Result:=str;
  bUtf:=IsUtf(str);
  for i:=128 to 255 do begin
    if(aChar[i]<>'') then begin
      if(bUtf) then aChar[i]:=AnsiToUtf(aChar[i]);
      Result:=ReplaceString(Result,aEntity[i],aChar[i]);
    end;
  end;
end;

////////////////////////////////////////////////////////////////////////////////
/// Ruzne metody hledani substringu
/// https://www.ncbi.nlm.nih.gov/pmc/articles/PMC61442/
//  !!! porovnava 8bitove znaky, tedy pokud je UTF, je treba upravit delky
//  !!! s ohledem na pocet vicebytovych znaku

/// http://members.upc.cz/nekup/3-algoritmy.htm
/// http://www.ms.mff.cuni.cz/~kopecky/vyuka/dis/02/dis02_v3.html

// Pro hledn slov bn dlky  (do 15 znaku) je nejlep algoritmus Boyer-Moore.
// Pro ast hledn slov extrmn dlky je vhodnj algoritmus Rabin-Karpa,
//    u nehoz temer delka vyhledavaneho vzorku neovlivnuje casovou narocnost zpracovani

{function Knuth_Morris_Pratt(const target, pattern: PChar; const lTarget, ;Pattern:integer):integer;
  //v target hleda pattern, lTarget a lPattern jsou delky stringu (protoze ty jsou zadane ukazatelem)
  //  http://martin.lipinsky.cz/skola/pt/HTML/63/text.htm
var step: array[0..255] of integer;
    i,j:integer;
begin
  Result:=-1;
  if(lTarget*lPattern=0) then exit; //neni co hledat, nektera z delek je nulova
  i:=0;
  j:=-1;
  step[0]:=-1;
  //priprava tabulky
  repeat
    if(j=-1) or (pattern[i]=pattern[j]) then begin
      Inc(i);
      Inc(j);
      if(pattern[j]=pattern[i] then step[i]:=step[j]
      else step[i]:=j;
    end
    else j:=step[j];
  until i=lPattern-1;
  j:=-1;
  i:=0;
  //vlastni smycka pro hledani substringu
  while i<lTarget do begin
    if(j=-1) or (pattern[j]=target[i]) then begin
      Inc(i);
      Inc(j);
      if(j>=lPattern then begin
        Result:=i-j+1;
        exit;
      end;
    end
    else j:=step[j];
  end;
end;
}
{
function Boyer_Moore_Horspool(const target,pattern:PChar; const lTarget,lPattern:integer):integer;
    //v target hleda pattern, lTarget a lPattern jsou delky stringu (protoze ty jsou zadane ukazatelem)
    // http://voho.eu/wiki/algoritmus-boyer-moore-horspool/
var i,j,k:integer;
    step:array[0..255] of integer;
begin
  Result:=0;
  if(lTarget*lPattern=0) then exit; //neni co hledat, nektera z delek je nulova
  //priprava tabulky
  for k:=0 to 255 do step[k]:=lPattern; //inicializace tabulky preskoku
  for k:=0 to lPattern do step[Ord(pattern[k])]:=lPattern-k; //tabulka preskoceni vzorku
  k:=lPattern;
  while(k<=lTarget) do begin
    i:=k-1;
    j:=lPattern-1;
    while target[i]=pattern[j] do begin
      Dec(i);
      Dec(j);
    end;
    if(j=-1) then begin
      Result:=i+2;
      exit;
    end;
    k:=k+step[Ord(target[k])];
  end;
end;
}

{
function Rabin_Karpa(const pat: string; const  Text: string) : integer;
  //https://stackoverflow.com/questions/30072555/rabin-karp-algorithm-in-pascal
  //ale tady se pouziva copy(), mozna je lepsi
  //http://www.algorytm.org/przetwarzanie-tekstu/algorytm-kr-karpa-rabina/kr-d.html
const
  b = 131;
var
  hpat, htext, Bm, j, m, n: integer;
  found : Boolean;
begin
  found := False;
  result := 0;
  m := length (pat);
  if m = 0 then  begin
    result := 1;
    found := true
  end;

  Bm := 1;
  hpat := 0;
  htext := 0;
  n := length (Text);
  if n >= m then begin   // preprocessing
    for j := 1 to m do begin
      Bm := Bm * b;
      hpat := hpat * b + ord (pat[j]);
      htext := htext * b + ord (Text[j])
    end;
  end;
  j := m;
  while not found do begin
    if (hpat = htext) and (pat = Copy (Text, j - m + 1, m)) then begin
      result := j - m;
      found := true
    end;
    if j < n then begin
      j := j + 1;
      htext := htext * b - ord (Text[j - m]) * Bm + ord (Text[j])
    end
    else found := true;
  end
end;
}